import sys
import os
import shutil

from IPython.display import (display, display_png, Image, FileLink)


#Constants
VERSION_PATH = "version.txt"
PS_FILE = "postScript.ps"
LATEX_FILE = "LaTeX.tex"
GENERIC_FOLDER = "Figure"
LATEX_CMD = "latex"
DVIPS_CMD = "dvips"
CONVERT_CMD = "convert_img"
WRAPPER = r"""
\documentclass{minimal}
\begin{document}
@REPLACE@
\end{document}
"""

#Global Variable
forceCompile = False


#--------------------------
#    Utility functions
#--------------------------


def clearVersion() :
    """Utility function used to clear the documents created by the compilation.
    It retrieves the current index for the compilation and remove all the previous folders.
    If the value stored for the index is incorrect, nothing is done.
    It also removes the file containing the current index.
    """
    if os.path.exists(VERSION_PATH) :
        f = open(VERSION_PATH, "r")
        value = str(f.read())
        f.close()
        if value.isdigit() :
            for i in range(int(value)) :
                shutil.rmtree("%s%i" % (GENERIC_FOLDER, i), True)
        os.remove(VERSION_PATH)

		
def loadVersion() :
    """Utility function loading the current index for the folder storing a compiled document.
    If the document storing the index is missing or the value contained is incorrect
    (typically, if it is a string) the index will be set to 0.
    Then, the index incremented is stored in the version file and the folder
    corresponding to the index is created.
    """
    value = ""
    if os.path.exists(VERSION_PATH) :
        f = open(VERSION_PATH, "r")
        value = f.read()
        f.close()
    else :
         f = open(VERSION_PATH, "w")
         f.write(str(0))
         f.close
    version = 0
    if value.isdigit() :
        version = int(value)
    f = open(VERSION_PATH, "w")
    f.write("%i" % (version+1))
    f.close()
    os.makedirs("%s%i" % (GENERIC_FOLDER, version))
    return version

	
def writeTeXFile(data, version=0) :
    """Write the data given in a TeX file corresponding to the current index.
    
    Keyword arguments:
    data -- the string containing the text to write in the document.
    version -- the index of the folder to store the file in (default 0).
    """
    path = os.path.join("%s%i" % (GENERIC_FOLDER, version), LATEX_FILE)
    f = open(path, 'w')
    f.write(data)
    f.close()
    return path
	

def writePSFile(data, version=0) :
    """Write the data given in a PS file corresponding to the current index.
    
    Keyword arguments:
    data -- the string containing the text to write in the document.
    version -- the index of the folder to store the file in (default 0).
    """
    path = os.path.join("%s%i" % (GENERIC_FOLDER, version), PS_FILE)
    f = open(path, 'w')
    f.write(data)
    f.close()
    return path


def convertLaTeXFile(path, doubleCompilation=False) :
    """Convert the TeX file at the specified path to a PS file,
    using the latex and dvips commands. If an error is encountered
    during the conversion, the process is stopped and an error message is displayed.
    Caution : this error message uses IPython functionalities.
    The double compilation is supported if needed. The intermediate files from the conversion are deleted.
	
    The process return the path to the PS file if the compilation was successful, and None otherwise.
    
    Keyword arguments:
    path -- the path of the TeX file.
    doubleCompilation -- boolean containing if a double compilation should be done (default False)
    """
    here = os.getcwd()
    filename = os.path.basename(path)[:-4]
    os.chdir(os.path.dirname(path))
    ret = os.system("%s -quiet %s" % (LATEX_CMD, filename + ".tex"))
    if ret != 0 :
        os.chdir(here)
        print("An error occurred during the conversion, please check the code. Details of the error in the following file :")
        display(FileLink(os.path.join(os.path.join(os.path.dirname(path), filename + ".log"))))
        return None
    if doubleCompilation :
        ret += os.system("%s -quiet %s" % (LATEX_CMD, filename + ".tex"))
    ret += os.system("%s -q %s -o %s" % (DVIPS_CMD, filename + ".dvi", filename + ".ps"))
    os.remove(filename + ".dvi")
    os.remove(filename + ".aux")
    os.remove(filename + ".log")
    os.chdir(here)
    if ret == 0 :
        return path[:-4] + ".ps"
    else :
        print("Something failed in the dvips conversion. Unexpected behaviour.")
        return None
	

def convertPSFile(path, pngPath=None) :
    """Convert the PS file at the specified path to one or multiple PNG files, one PNG image per page.
    The path to save the PNG can be specified, is must be a file and not a directory.
    The conversion uses a command from ImageMagick, convert_img (it is a renamed version of convert).
    If no PNG path is specified, the name and the emplacement of the postScript file will be used instead.
    If a path different from a PNG file is given, a missing extension will be assumed.
    This function returns the path of the PNG image (even if multiple images will have distinct paths).
	
    Keyword arguments:
    path : the path of the PS file.
    pngPath : the path where to save the PNG file(s) (default None)
    """
    if pngPath is None :
        pngPath = path[:-3] + ".png"
    else :
        if len(pngPath)<4 or (pngPath[-4:] != ".png" and pngPath[-4:] != ".PNG"):
            print("The path given was not a png file. Assuming missing extension.")
            pngPath = pngPath + ".png"
    os.system("%s %s %s" % (CONVERT_CMD, path, pngPath))
    return pngPath


def findPNGFiles(pngPath, expectations=False) :
    """Utility function to retrieve one or multiple PNG files from an emplacement.
    The given path can design a file or the prefix of multiple PNG files
    (in the format prefix-x.png, where x is an integer). If a file is found, the prefix case
    won't be evaluated.
    All the path found are stored in a table and returned by the function.
    If the table is empty and it was expected to find results, an error message is displayed.
	
    Keyword arguments:
    pngPath -- the file's emplacement or the common prefix.
    expectations -- if the function displays an error message in case of empty result (default False).
    """
    result = []
    if len(pngPath)<4 or pngPath[-4:] != ".png" :
        print("The path given was not a png file. Assuming missing extension.")
        pngPath = pngPath + ".png"
    if os.path.exists(pngPath) :
        result.append(pngPath)
    else :
        pngPath = pngPath[:-4] + "-"
        stillSeeking = True
        index = 0
        while(stillSeeking) :
            currentPath = "%s%i.png" % (pngPath, index)
            if os.path.exists(currentPath) :
                result.append(currentPath)
                index = index + 1
            else :
                stillSeeking = False
    if expectations and len(result)<1 :
        print("Conversion in PNG file didn't work, please check the postScript code you provided.")
    return result


#--------------------------
#    Mid-level functions
#--------------------------


def displayPScode(data, pngPath=None) :
    """This function loads and displays the PNG file specified in argument,
    unless the forced compilation is enabled. If no PNG file can be displayed
    or if the compilation is forced, it writes the given text in an PS file and
    calls displayPSFile on it, transmitting the PNG path.
	
    Keyword arguments:
    data -- the string containing the PS code.
    pngPath -- the path where to fetch/save the PNG file (default None).
    """
    if pngPath is not None and not forceCompile :
        if checkPNGFiles(pngPath) :
            return None
    folder = 0
    if pngPath is None :
        folder = loadVersion()
    path = writePSFile(data, folder)
    displayPSFile(path, pngPath)
    return None


def displayPSFile(path, pngPath=None) :
    """This function loads and displays the PNG file specified in argument,
    unless the forced compilation is enabled. If no PNG file can be displayed
    or if the compilation is forced, it converts the PS file given in argument
    and displays the images resulting of the conversion.
	
    Keyword arguments:
    path -- the path of the PS file to convert.
    pngPath -- the path where to fetch/save the PNG file (default None).
    """
    if pngPath is not None and not forceCompile :
        if checkPNGFiles(pngPath) :
            return None
    result = convertPSFile(path, pngPath)
    checkPNGFiles(result, True)
    return None


def displayLatexCode(data, pngPath=None, doubleCompilation=False) :
    """This function loads and displays the PNG file specified in argument,
    unless the forced compilation is enabled. If no PNG file can be displayed
    or if the compilation is forced, it writes the given text in an TeX file and
    calls displayLatexFile on it, transmitting the PNG path and the indication for
    the double compilation.
	
    Keyword arguments:
    data -- the string containing the LaTeX code.
    pngPath -- the path where to fetch/save the PNG file (default None).
    doubleCompilation -- whether the compilation should be executed twice (default False)
    """
    if pngPath is not None and not forceCompile :
        if checkPNGFiles(pngPath) :
            return None
    folder = 0
    if pngPath is None :
        folder = loadVersion()
    path = writeTeXFile(data, folder)
    displayLatexFile(path, pngPath, doubleCompilation)
    return None


def displayLatexFile(path, pngPath=None, doubleCompilation=False) :
    """This function loads and displays the PNG file specified in argument,
    unless the forced compilation is enabled. If no PNG file can be displayed
    or if the compilation is forced, it converts the LaTeX file given in argument,
    using the double compilation if specified, generating a PS file. The result is then
    converted and displayed using displayPSFile.
	
    Keyword arguments:
    path -- the path of the LaTeX file to convert.
    pngPath -- the path where to fetch/save the PNG file (default None).
    doubleCompilation -- whether the compilation should be executed twice (default False)
    """
    if pngPath is not None and not forceCompile :
        if checkPNGFiles(pngPath) :
            return None
    result = convertLaTeXFile(path, doubleCompilation)
    if result is not None :
        displayPSFile(result, pngPath)
    return None


def checkPNGFiles(path, expectations=False) :
    """Retrieve and display the PNG images corresponding to the specified path.
    If no file is found can display an error message.
    The algorithm used for retrieving files is given in findPNGFiles.
	
    Keyword arguments:
    path -- the path of the PNG file or the common prefix of multiple files.
    expectations -- if the function displays an error message in case of empty result (default False).
    """
    result = findPNGFiles(path, expectations)
    if len(result) > 0 :
        for elem in result :
            display_png(Image(filename=elem))
        return True
    else :
        return False

	
#--------------------------
#    Magics functions
#--------------------------


def PSFile(line) :
    """The function defining the %PSFile magic.
    The expected arguments are the path to the PS file to display and optionally
    the path of the PNG file corresponding (if not found or compilation forced, it
    will be the converted file's path).
    In case of incorrect use, this function displays an error message.
	
    Keyword arguments:
    line -- the line of text after the magic's invocation, of the form path [path]
    """
    paths = line.split(" ")
    if len(paths) > 2 or len(paths) == 0 :
        print("Incorrect use of the magic. Please see '%PSViewer -help' for help")
    else :
        if len(paths) == 1 :
            displayPSFile(paths[0])
        else :
            displayPSFile(paths[0], paths[1])


def LatexFile(line) :
    """The function defining the %LatexFile magic.
    The expected arguments (the line) are the path to the Latex file to display and optionally
    the path of the PNG file corresponding (if not found or compilation forced, it
    will be the converted file's path).
    In case of incorrect use, this function displays an error message.
	
    Keyword arguments:
    line -- the line of text after the magic's invocation, of the form path [path]
    """
    paths = line.split(" ")
    if len(paths) > 2 or len(paths) == 0 :
        print("Incorrect use of the magic. Please see '%PSViewer -help' for help")
    else :
        if len(paths) == 1 :
            displayLatexFile(paths[0])
        else :
            displayLatexFile(paths[0], paths[1])


def PSCode(line, cell) :
    """The function defining the %%PSCode magic.
    The line argument contains the options of the magic. If a path is specified as
    first option, the magic will try to retrieve PNG images from this emplacement
    and display them without any compilation. If the compilation is forced or if no
    file are found, the following options will be evaluated and the conversion of the
    cell's code will be execute.
    The current options are :
    -noSave which allow the override of the intermediate files. Caution : this can cause
    unexpected behaviour in some cases.
    Any other command will be ignored.
	
    Keyword arguments:
    line -- the line of text after the magic's invocation, of the form "path [option*]"
    cell -- the content of the cell, expected to be the postScript code.
    """
    if len(line) > 1 :
        commands = line.split(" ")
        path = None
        if commands[0][0] != "-" :
            path = commands[0]
        if "-noSave" in commands and pngPath is None :
            path = os.path.join(GENERIC_FOLDER + "0", "display.png")
        displayPScode(cell, path)
    else :
        displayPScode(cell)

def LatexCode(line, cell) :
    """The function defining the %%LatexCode magic.
    The line argument contains the options of the magic. If a path is specified as
    first option, the magic will try to retrieve PNG images from this emplacement
    and display them without any compilation. If the compilation is forced or if no
    file are found, the following options will be evaluated and the conversion of the
    cell's code will be execute.
    The current options are :
    -noSave which allow the override of the intermediate files. Caution : this can cause
    unexpected behaviour in some cases.
    -noImport generates the documentclass, begin{document} and end{document} of the
    file automatically.
    -doubleCompilation indicates that a double compilation is required. This option is useful
    in case of internal links.
    Any other command will be ignored.
	
    Keyword arguments:
    line -- the line of text after the magic's invocation, of the form "path [option*]"
    cell -- the content of the cell, expected to be the LaTeX code.
    """
    if len(line) < 1 :
        displayLatexCode(cell)
    else :
        commands = line.split(" ")
        pngPath = None
        if commands[0][0] != "-" :
            pngPath = commands[0]
        if "-noImport" in commands :
            cell = WRAPPER.replace("@REPLACE@", cell)
        noSave = None
        doubleCompilation = False
        if "-noSave" in commands and pngPath is None :
            noSave = os.path.join(GENERIC_FOLDER + "0", "display.png")
        if "-doubleCompilation" in commands :
            doubleCompilation = True
        if noSave is not None or doubleCompilation :
            displayLatexCode(cell, noSave, doubleCompilation)
        else :
            displayLatexCode(cell)

def PSViewer(line) :
    """The function defining the %PSViewer magic.
    It mainly contains utility functionalities in order to modify slightly the
    behaviour of the module or retrieve some informations.
    The magic is expected to be followed by an option from: 
    -reset -- delete the files created during the compilations, expect if they have
    been stored in a specified path.
    -path -- display the path to the folder where the intermediate files are created.
    -forceCompile -- toggle the forced compilation of the code/files given.
    -compilationState -- display if the compilation is forced or not.
    Any other option (even no option) will be interpreted as -help and will display the help.
	
    Keyword arguments:
    line -- the line of text after the magic's invocation, of the form "option"
    """
    import PSViewer
    if line == "-reset" :
        clearVersion()
        print ("Version cleared, all files deleted")
    elif line == "-path" :
        print (os.getcwd())
    elif line == "-forceCompile" :
        if PSViewer.forceCompile :
            PSViewer.forceCompile = False
            print("Toggled behaviour to displaying given image")
        else :
            PSViewer.forceCompile = True
            print("Toggled behaviour to forced compilation")
    elif line == "-compilationState" :
        if PSViewer.forceCompile :
            print("Default behaviour : forced compilation")
        else :
            print("Default behaviour : displaying given image")
    else :
        print (r"""
The following magics are defined in this module :
	(all the options must be on the same line that the magic)

%%PSCode allows to run postScript code directly in the cell.
	Giving a path as first argument will modify the behaviour of the magic.
	If one or multiple png images are found at this emplacement, they will be displayed instead of the compilation. In the other case or if forceCompile is enabled, the file will be compiled and displayed. 
	the option -noSave prevent the excessive usage of memory by overriding the intermediate files in memory.
	Caution : this option can have unexpected result if you convert a document with less pages than the previous one.

%PSFile allows to view the content of a ps file in the notebook.
	The first argument is always the path to the file. A second path can be given. With default behaviour, the png images corresponding to the path will be loaded and displayed. With forceCompile or if the path is empty, the png generated by the magic will be stored at this emplacement.
	The element(s) must be in absolute path or in the notebook's directory.
	
%%LatexCode allows to run LaTeX code in the cell. Following options exist :
	Giving a path as first argument will modify the behaviour of the magic. If one or multiple png images are found at this emplacement, they will be displayed instead of the compilation. In the other case or if forceCompile is enabled, the file will be compiled and displayed.
	-noSave prevent the excessive usage of memory by overriding the intermediate files created in memory.
	Caution : this option can have unexpected result if you convert a document with less pages than the previous one.
	-doubleCompilation is useful if you need to compile a LaTeX document with references in it.
	You cannot compile multiple times just by reinterpreting the cell.
	-noImport allow to write a basic LaTeX document without import, without specifying \documentclass, \begin{document} and \end{document}
	
%LatexFile allows to view the content of a LaTeX file in the notebook.
	The first argument is always the path to the file. A second path can be given. With default behaviour, the png images corresponding to the path will be loaded and displayed. With forceCompile or if the path is empty, the png generated by the magic will be stored at this emplacement.
	The element(s) must be in absolute path or in the notebook's directory.
	
%PSViewer is the set of utility functions. It is mostly for help.
	-reset clear all the files created with the other magics. It can't be undone.
	-path show the path to the default folder containing the files created with the magics.
	-forceCompile toggle the forced compilation of the LaTeX/PS code.
	-compilationState displays what is the treatment for the path given for the conversion.
	Any other argument will be interpreted as -help and will display this help.""")


#--------------------------
#    Declaring magics
#--------------------------


def load_ipython_extension(ipython) :
    """The function defining the magics and which is their type."""
    ipython.register_magic_function(PSViewer, "line")
    ipython.register_magic_function(PSFile, "line")
    ipython.register_magic_function(LatexFile, "line")
    ipython.register_magic_function(PSCode, "cell")
    ipython.register_magic_function(LatexCode, "cell")